from __future__ import annotations
from typing import Iterable, Union
from enum import IntEnum

NetworkAttributeValue = Union[int, float, bool]
NetworkAttributeId = int
NetworkAttributeParameter = tuple[str, Union[float, None], NetworkAttributeParameterUsageType]
NetworkElementId = int
NetworkSourceId = int
NetworkSourceObjectId = int
FromPosition = float
ToPosition = float
FromAzimuthValue = float
ToAzimuthValue = float
OleDateTime = float

class NetworkTimeUsage(IntEnum):
    '''Describes various time usage types.'''

    BeforeTraversal = 1
    AfterTraversal = 2

class NetworkAttributeUsageType(IntEnum):
    '''Describes various network attribute usage types.'''

    Cost = 0
    Descriptor = 1
    Restriction = 2
    Hierarchy = 3

class NetworkAttributeParameterUsageType(IntEnum):
    '''Describes various network attribute parameter usage types.'''

    General = 0
    Restriction = 1

class NetworkEdgeDirection(IntEnum):
    '''Describes various network edge directions.'''
    Undefined = 0
    Along = 1
    Against = 2

class NetworkDatasetQuery:
    '''Provides the access to the network dataset attributes and elements.'''

    def __init__(self, in_network: str):
        '''
        Initialize network query with the catalog path to the network dataset,
        a network dataset layer object, or the string name of the network dataset layer.
        '''
        ...

    def queryNetworkName(self) -> str:
        '''Returns the name of the network.'''
        ...

    def querySourceNames(self) -> list[str]:
        '''Returns the list of network source names.'''
        ...

    def queryAttributeId(self, attributeName: str) -> NetworkAttributeId:
        '''Returns the atribute id by the attribute name.'''
        ...

    def queryAttributeUsageType(self, id: NetworkAttributeId) -> NetworkAttributeUsageType:
        '''Returns the network attribute usage type.'''
        ...

    def queryAttributeParameters(self, id: NetworkAttributeId) -> Iterable[NetworkAttributeParameter]:
        '''Returns the network attribute parameters.'''
        ...

    def queryJunctions(self, edgeId: int) -> tuple[NetworkElementId, NetworkElementId]:
        '''Returns 'from' and 'to' junctions for the given edge.'''
        ...

    def queryAtJunction(self, turnId: NetworkElementId) -> NetworkElementId:
        '''Returns the junction that connects first and second edges of the turn element.'''
        ...

    def queryTurns(self, junctionId: NetworkElementId) -> Iterable[NetworkElementId]:
        '''Returns all turn elements for the given junction that connects first and second edges of each turn element.'''
        ...

    def queryEdges(self, junctionId: NetworkElementId) -> Iterable[NetworkElementId]:
        '''Returns all edges that are connected using specified junction.'''
        ...

    def queryTurnEdges(self, turnId: NetworkElementId) -> Iterable[NetworkElementId]:
        '''Returns all edges that describe specified turn element.'''
        ...

    def queryEdgePositions(self, edgeId: NetworkElementId) -> tuple[FromPosition, ToPosition]:
        '''Returns 'from' and 'to' edge positions on the network source feature.'''
        ...

    def queryEdgeAzimuths(self, edgeId: NetworkElementId) -> tuple[FromAzimuthValue, ToAzimuthValue]:
        '''Returns 'from' and 'to' azimuth values for the given edge element.'''
        ...

    def queryJunctionSourceInfo(self, junctionId: NetworkElementId) -> tuple[NetworkSourceId, NetworkSourceObjectId]:
        '''Returns the junction element source information in the of network source ID and network source object ID.'''
        ...

    def queryEdgeSourceInfo(self, edgeId: NetworkElementId) -> tuple[NetworkSourceId, NetworkSourceObjectId]:
        '''Returns the edge element source information in the of network source ID and network source object ID.'''
        ...

    def queryTurnSourceInfo(self, turnId: NetworkElementId) -> tuple[NetworkSourceId, NetworkSourceObjectId]:
        '''Returns the turn element source information in the of network source ID and network source object ID.'''
        ...

    def queryJunctionAttributeValue(self, junctionId: NetworkElementId, attributeId: NetworkAttributeId) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    def queryEdgeAttributeValue(self, edgeId: NetworkElementId, direction: NetworkEdgeDirection, attributeId: NetworkAttributeId) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    def queryTurnAttributeValue(self, turnId: NetworkElementId, attributeId: NetworkAttributeId) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    def queryJunctionAttributeValueAtTime(self, junctionId: NetworkElementId, attributeId: int, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    def queryEdgeAttributeValueAtTime(self, edgeId: NetworkElementId, direction: NetworkEdgeDirection, attributeId: int, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    def queryTurnAttributeValueAtTime(self, turnId: NetworkElementId, attributeId: int, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the attribute value for the given attribute.'''
        ...

    ...

class NetworkAttributeEvaluator:
    
    def __init__(self, attr_name: str, supported_sources: list[str] = None) -> None:
        '''Constructs network attribute evaluator.'''
        ...

    @property
    def attributeName(self) -> str:
        '''Returns the name of network attribute that is going to be intercepted.'''
        ...

    @property
    def sourceNames(self) -> list[str]:
        '''Returns the list of assigned network source names.'''
        ...
    
    @property
    def networkQuery(self) -> NetworkDatasetQuery:
        '''Returns the instance to the attached network dataset query or None if the evaluator was not attached yet.'''
        ...

    def attach(self, network_query: NetworkDatasetQuery) -> bool:
        ''' Will be called once to when attaching to the network dataset.'''
        ...

    def refresh(self) -> None:
        ''' Will be called every time when internal cache is reset.'''
        ...

    ...

class NetworkEdgeAttributeEvaluator(NetworkAttributeEvaluator):

    def __init__(self, attr_name: str, supported_sources: list[str] = None) -> None:
        '''Constructs network edge attribute evaluator.'''
        ...

    def queryAttributeValue(self, eid: NetworkElementId, direction: NetworkEdgeDirection) -> NetworkAttributeValue:
        '''Returns the calculated network edge attribute value.'''
        ...

    def queryAttributeValueAtTime(self, eid: NetworkElementId, direction: NetworkEdgeDirection, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the calcualte network attribute values for the given network edge element and time.'''
        ...

    ...

class NetworkJunctionAttributeEvaluator(NetworkAttributeEvaluator):

    def __init__(self, attr_name: str, supported_sources: list[str] = None) -> None:
        '''Constructs network junction attribute evaluator.'''
        ...

    def queryAttributeValue(self, eid: NetworkElementId) -> NetworkAttributeValue:
        '''Returns the calculated network junction attribute value.'''
        ...

    def queryAttributeValueAtTime(self, eid: NetworkElementId, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the calcualte network attribute values for the given network junction element and time.'''
        ...

    ...

class NetworkTurnAttributeEvaluator(NetworkAttributeEvaluator):

    def __init__(self, attr_name: str, supported_sources: list[str] = None) -> None:
        '''Constructs network turn attribute evaluator.'''
        ...

    def queryAttributeValue(self, eid: NetworkElementId) -> NetworkAttributeValue:
        '''Returns the calculated network turn attribute value.'''
        ...

    def queryAttributeValueAtTime(self, eid: NetworkElementId, time: OleDateTime, time_usage: NetworkTimeUsage) -> NetworkAttributeValue:
        '''Returns the calcualte network attribute values for the given network turn element and time.'''
        ...

    ...
