import datetime
from __future__ import annotations
from enum import IntEnum
from typing import Any, Union, Iterable
from ._nds import NetworkDatasetQuery
import abc

JunctionCode = 1 << 16
EdgeCode = 1 << 17
TurnCode = 1 << 18

NetworkAttributeValue = Union[int, float, bool]
# ! date time will be handled by float because we don't want conversions
NetworkSourceFieldValue = Union[int, float, str]
NetworkElementId = int


class TraversedElementType(IntEnum):
    # fall back type
    Unknown = 0,

    # basic network elements
    Junction = JunctionCode,
    Edge = EdgeCode,
    Turn = TurnCode,

    # edges (streets)
    Road = EdgeCode + 1,
    HighwayRoad = EdgeCode + 2,
    Ramp = EdgeCode + 3,
    Ferry = EdgeCode + 4,
    RoundaboutRoad = EdgeCode + 5,
    MajorRoad = EdgeCode + 6,

    # edges (campus)
    Walkway = EdgeCode + 10,
    TurningArc = EdgeCode + 11,
    Stairs = EdgeCode + 12,
    Escalator = EdgeCode + 13,
    Elevator = EdgeCode + 14,
    PedestrianRamp = EdgeCode + 15,
    MovingWalkway = EdgeCode + 16,
    Hallway = EdgeCode + 17,
    Indoor = EdgeCode + 18,

    # edges (transit)
    Transit = EdgeCode + 64,

    # edges (maritime)
    SailingLine = EdgeCode + 128,

    # junctions
    Stop = JunctionCode + 100,
    Waypoint = JunctionCode + 101,
    RestBreak = JunctionCode + 102,
    RoadIntersection = JunctionCode + 103,

class CurbApproach(IntEnum):
    RightSide = 1,
    LeftSide = 2

class DrivingSide(IntEnum):
    Unknown = 0,
    RightSide = 1,
    LeftSide = 2

class LandmarkSide(IntEnum):
    Both = 0,
    Left = 1,
    Right = 2

class ReferenceLandmarkType(IntEnum):
    Turn = 0,
    Confirmation = 1,
    StopSign = 2,
    TrafficLight = 3,
    RailwayCrossing = 4


class NameClass(IntEnum):
    # TODO: Is NameClass too generic? Alternative name?
    StreetName = 0,
    RouteNumber = 1,


class DirectionPointType(IntEnum):
    Unknown = 0,
    Header = 1,
    ManeuverArrive = 50,
    ManeuverDepart = 51,
    ManeuverStraight = 52,
    ManeuverFerryOn = 100,
    ManeuverFerryOff = 101,
    ManeuverForkCentral = 102,
    ManeuverRoundabout = 103,
    ManeuverUTurn = 104,
    ManeuverDoor = 150,
    ManeuverStairs = 151,
    ManeuverElevator = 152,
    ManeuverEscalator = 153,
    ManeuverPedestrianRamp = 154,
    ManeuverForkLeft = 200,
    ManeuverRampLeft = 201,
    ManeuverRoundaboutLeft = 202,
    ManeuverUTurnLeft = 203,
    ManeuverBearLeft = 204,
    ManeuverTurnLeft = 205,
    ManeuverSharpLeft = 206,
    ManeuverTurnLeftLeft = 207,
    ManeuverTurnLeftRight = 208,
    ManeuverForkRight = 300,
    ManeuverRampRight = 301,
    ManeuverRoundaboutRight = 302,
    ManeuverUTurnRight = 303,
    ManeuverBearRight = 304,
    ManeuverTurnRight = 305,
    ManeuverSharpRight = 306,
    ManeuverTurnRightLeft = 307,
    ManeuverTurnRightRight = 308,
    ManeuverElevatorUp = 400,
    ManeuverEscalatorUp = 401,
    ManeuverStairsUp = 402,
    ManeuverElevatorDown = 500,
    ManeuverEscalatorDown = 501,
    ManeuverStairsDown = 502,
    Event = 1000,
    EventLandmark = 1001,
    EventTimeZone = 1002,
    EventTraffic = 1003,
    EventBarrier = 1004,
    EventBoundary = 1005,
    EventRestrictionViolation = 1006


class DirectionPoint:
    """
    Creatable class that represents maneuver instruction.
    Creates with ref@RouteEdge or ref@RouteJunction
    """

    @property
    def displayText(self) -> str:
        """The directions text, formatted to be used for UI display."""
        pass

    @property
    def voiceInstruction(self) -> str:
        """The directions text, formatted to be used for voice guidance.  The text will have expanded abbreviations and
        plurals."""
        pass

    @property
    def shortVoiceInstruction(self) -> str:
        """A brief version of the directions text, formatted to be used for voice guidance.  The text will have
        expanded abbreviations and plurals."""
        pass

    @property
    def directionPointType(self) -> DirectionPointType:
        """The direction point type."""
        pass

    @property
    def positionAlong(self) -> float:
        """
        The position on route edge between ref@RouteEdge.fromPosition and ref@RouteEdge.toPosition.
        If the direction point is on a junction the value should remain at the default value (0.0).
        """
        pass

    @property
    def azimuth(self) -> float:
        """
        The bearing, in degrees, of the vehicle departing this point in range [0, 360]. Zero indicates north.
        It will be regenerated by the engine for new points.
        """
        pass

    @property
    def arrivalTime(self) -> datetime:
        """
        The time when the directions event happens in coordinated universal time (UTC).
        It will be regenerated by the engine for new points.
        """
        pass

    @property
    def arrivalUTCOffset(self) -> float:
        """
        The difference, in minutes, between the local arrival time and the UTC arrival time.
        """
        pass

    @property
    def exitNumber(self) -> int:
        """The exit number on a roundabout."""
        pass

    @property
    def exitName(self) -> SourceFeatureName:
        """The highway exit name that appears in the directions text."""
        pass

    @property
    def name(self) -> SourceFeatureName | None:
        """The name of the directions point."""
        pass

    @property
    def alternateName(self) -> SourceFeatureName | None:
        """Alternate source name that appears in the directions text."""
        pass

    @property
    def intersectingNames(self) -> list[SourceFeatureName]:
        """The names of the intersecting or cross street that appears in the directions text."""
        pass

    @property
    def branchNames(self) -> list[SourceFeatureName]:
        """The signpost branch names that appears in the directions text."""
        pass

    @property
    def towardNames(self) -> list[SourceFeatureName]:
        """The signpost toward destination names that appears in the directions text."""
        pass

class SourceFeatureNameType(IntEnum):
    # TODO: Alternative name for SourceFeatureName: DirectionsName, DirectionsFeatureName
    PrimaryName = 0
    AlternateName1 = 1
    AlternateName2 = 2
    AlternateName3 = 3
    AlternateName4 = 4
    AlternateName5 = 5


class SourceFeatureNameField(IntEnum):
    # TODO: I am not very attached to the name. Do we have better options for the name?
    FullName = 0
    BaseName = 1
    PrefixType = 2
    SuffixType = 3
    HighwayDirection = 4
    PrefixDirection = 5
    SuffixDirection = 6
    Phrase = 7
    Language = 8
    NameClass = 9 # not implemented


class DirectionsCustomizer(abc.ABC):
    """Base class for directions customizers."""
    @abc.abstractmethod
    def attach(self, ndv: NetworkDatasetQuery) -> bool:
        """ Will be called once when attaching to the network dataset."""
        pass

    @abc.abstractmethod
    def customize(self, trv: DirectionsResultQuery) -> None:
        """
        The method must be overridden in a derived class, it does nothing by default.
        It will be called by the directions engine once per route.
        
        Below is a example that prefixes the directions texts with the word ALTERED:
        class SampleAlterText(arcpy.nax.DirectionsCustomizer):
            def customize(self, query):
                for junction in query.junctions:
                    directionPoints = query.queryDirectionPoints(junction)
                    for point in directionPoints:
                        point.displayText = "ALTERED " + point.displayText
                    query.updateDirectionPoints(junction, directionPoints)
        """
        pass


class DirectionsResultQuery:
    """TODO: doc"""

    @property
    def routeId(self) -> int:
        """The zero-based ID of the route in the result."""
        pass

    @property
    def routeName(self) -> str:
        """The name of the route."""
        pass

    @property
    def softRestrictionNames(self) -> list[str]:
        """The list of enabled soft restrictions."""
        pass

    # region: Traversibility
    # Iterate
    @property
    def junctions(self) -> Iterable[TraversedJunction] | None:
        """Return all the junction elements in the sequence being traversed."""
        pass

    @property
    def edges(self) -> Iterable[TraversedEdge] | None:
        """Return all the edge elements in the sequence being traversed."""
        pass

    #@property
    #def turns(self) -> Iterable[TraversedTurn] | None:
    #    """Return all the turn elements in the sequence being traversed."""
    #    pass

    # Traverse
    def previousTraversedEdge(self, junction: TraversedJunction) -> TraversedEdge | None:
        """Return previous traversed edge element of provided junction."""
        pass

    def nextTraversedEdge(self, junction: TraversedJunction) -> TraversedEdge | None:
        """Return next traversed edge element of provided junction."""
        pass

    def adjacentEdges(self, junction: TraversedJunction) -> Iterable[AdjacentNetworkEdge] | None:
        """Return adjacent edges of provided junction element. These edges are not traversed."""

    def queryTurns(self, junction: TraversedJunction) -> Iterable[TraversedTurn] | None:
        """Return turns that is taken at a particular junction."""
        pass

    def fromJunction(self, edge: TraversedEdge) -> TraversedJunction:
        """Return from junction of provided edge element."""
        pass

    def toJunction(self, edge: TraversedEdge) -> TraversedJunction:
        """Return to junction of provided edge element."""
        pass

    def fromEdge(self, turn: TraversedTurn) -> TraversedEdge:
        """Return from edge of provided turn element."""
        pass

    def toEdge(self, turn: TraversedTurn) -> TraversedEdge:
        """Return to edge of provided turn element."""
        pass
    # endregion

    # region: Directions access
    def queryDirectionPoints(self, element: TraversedElement) -> list[DirectionPoint]:
        """TODO: doc"""
        pass
    
    def updateDirectionPoints(self, element: TraversedElement, directions: list[DirectionPoint]) -> None:
        """TODO: doc"""
        pass
    # endregion

    # region: Query traversed element information
    def queryNetworkSourceFieldValue(self, element: TraversedElement, mapped_field_name: str) -> NetworkSourceFieldValue:
        """
        Returns the value of the specified mapped field (as defined in the Field Mappings of the network's directions
        properties) for the edge.
        None is returned if the field doesn't exist.
        """
        pass

    def queryNetworkAttributeValue(self, element: TraversedElement, attribute_id: int) -> NetworkAttributeValue:
        """
        Returns the value of the specified network attribute for the element.
        None is returned if the attribute doesn't exist.
        """
        pass
    # endregion


class TraversedElement:
    """Represents a single traversed element.
    Used as a base class for shared properties of TraversedEdge, TraversedJunction and TraversedTurn."""
    @property
    def type(self) -> TraversedElementType:
        """The type of the element."""
        pass

    @property
    def eid(self) -> int:
        """The element ID of element's network element."""
        pass

    @property
    def sid(self) -> int:
        """The source ID of element's network element."""
        pass

    @property
    def soid(self) -> int:
        """The source object ID of element's network element."""
        pass

    @property
    def rowID(self) -> int:
        """
        The traversed result element object ID.
        """
        pass

    @property
    def routePartIndex(self) -> int:
        """
        The index number of a route part from one stop to another, starting at zero.
        For example, if there is a route with stops Stop1, Stop2, Stop3, the 0 route
        part means the current edge lies between Stop1 and Stop2.
        """
        pass

    @property
    def accumulatedTime(self) -> float:
        """Accumulated time of the route up to and including this element. Expressed in the network's time attribute
        units."""
        pass

    @property
    def accumulatedLength(self) -> float:
        """Accumulated length of a route up to and including this element. Expressed in the network's length attribute units."""
        pass

    @property
    def accumulatedImpedance(self) -> float:
        """Accumulated impedance of a route up to and including this element. Expressed in the network's impedance attribute units."""
        pass

    @property
    def violatedRestrictions(self) -> list[str]:
        """The list of violated soft restrictions on the current element. The subset of ref@DirectionsResultView.softRestrictionNames."""
        pass


class TraversedTurn(TraversedElement):
    """Represents a single turn of a route."""
    # No special properties for turns for now, just whatever is on TraversedElement.
    pass


class TraversedJunction(TraversedElement):
    """
    Represents a single junction of a route.
    A typical route is an alternating sequence of edges and junctions, for example:
    RouteJunction -> RouteEdge -> RouteJunction -> RouteEdge -> RouteJunction
    """
    @property
    def directionPoints(self) -> list[DirectionPoint]:
        """The list of directions that refers to this route junction."""
        pass

    @property
    def violationTime(self) -> float:
        """The violation time. Expressed in the network's time attribute units."""
        pass

    @property
    def signpostID(self) -> int:
        """The OID of related signpost."""
        pass

    @property
    def roadSplitID(self) -> int:
        """The OID of related roadsplit."""
        pass

    @property
    def names(self) -> list[SourceFeatureName]:
        """The names of the junction."""
        pass

    @property
    def referenceLandmark(self) -> ReferenceLandmark | None:
        """The reference landmark related to the junction."""
        pass

    @property
    def spatialLandmarks(self) -> list[SpatialLandmark]:
        """The list of spatial landmarks related to the junction."""
        pass

    @property
    def level(self) -> int | None:
        """The level of the junction."""
        pass

    @property
    def timeWindowStart(self) -> datetime | None:
        """The start value of a time window in which a stop on the junction will be visited (in UTC)."""
        pass

    @property
    def timeWindowEnd(self) -> datetime | None:
        """The end value of a time window in which a stop on the junction will be visited (in UTC)."""
        pass

    @property
    def waitTime(self) -> float | None:
        """The wait time between the arrival and the departure for a stop.  Expressed in the network's time attribute
        units."""
        pass

    @property
    def utcOffset(self) -> float | None:
        """The offset from UTC in minutes."""
        pass

    @property
    def arrivalTime(self) -> datetime | None:
        """The estimated time of arrival (in UTC)."""
        pass

    @property
    def departureTime(self) -> datetime | None:
        """The estimated time of departure (in UTC)."""
        pass

    @property
    def arriveCurbApproach(self) -> CurbApproach | None:
        """The curb approach of a destination (if any) in junction."""
        pass

    def getName(self, index: SourceFeatureNameType, name_field: SourceFeatureNameField = SourceFeatureNameField.FullName, fallback_value="") -> str | None:
        """TODO: doc"""
        pass


class TraversedEdge(TraversedElement):
    """
    Represents a single edge of a route.
    A typical route is an alternating sequence of edges and junctions, for example:
    RouteJunction -> RouteEdge -> RouteJunction -> RouteEdge -> RouteJunction
    """

    @property
    def fromPosition(self) -> float:
        """The position on a network element where the route edge starts.  In the range of 0.0 to 1.0."""
        pass

    @property
    def toPosition(self) -> float:
        """The position on a network element where the route edge ends.  In the range of 0.0 to 1.0."""
        pass

    @property
    def immediateAzimuths(self) -> tuple[float, float]:
        """The azimuth calculated at close proximity from edge start.
        Return a tuple of "from" and "to" Azimuth"""
        pass

    @property
    def generalizedAzimuths(self) -> tuple[float, float]:
        """More generalized azimuth calculated further away from edge start."""
        pass

    @property
    def names(self) -> list[SourceFeatureName]:
        """The names of the edge."""
        pass

    @property
    def referenceLandmarks(self) -> list[ReferenceLandmark]:
        """The list of reference landmarks related to the edge."""
        pass

    @property
    def spatialLandmarks(self) -> list[SpatialLandmark]:
        """The list of spatial landmarks related to the edge."""
        pass

    @property
    def drivingSide(self) -> DrivingSide | None:
        """The driving side."""
        pass

    @property
    def fromLevel(self) -> int | None:
        """From which level the edge starts."""
        pass

    @property
    def toLevel(self) -> int | None:
        """On which level the edge ends."""
        pass

    @property
    def administrativeArea(self) -> str | None:
        """The name of administrative area the edge placed on."""
        pass

    @property
    def floorName(self) -> str | None:
        """The name of a floor the edge placed on."""
        pass

    @property
    def utcOffset(self) -> float | None:
        """The offset from UTC in minutes."""
        pass

    @property
    def timezoneId(self) -> int | None:
        """The ID of the time zone the edge placed on."""
        pass

    @property
    def arrivalTime(self) -> datetime | None:
        """The ETA of the start of the edge (in UTC)."""
        pass

    @property
    def departureTime(self) -> datetime | None:
        """The ETA of the end of the edge (in UTC)."""
        pass

    def getName(self, index: SourceFeatureNameType, name_field: SourceFeatureNameField = SourceFeatureNameField.FullName, fallback_value="") -> str | None:
        """TODO: doc"""
        pass

class AdjacentNetworkEdge:
    """
    Represents an edge adjacent to a junction.
    """

    @property
    def type(self) -> TraversedElementType:
        """The type of adjacent edge."""
        pass

    @property
    def immediateAzimuths(self) -> tuple[float, float]:
        """The azimuth calculated at close proximity from edge start."""
        pass

    @property
    def eid(self) -> int:
        """The element ID of adjacent edge's network element."""
        pass

    @property
    def sid(self) -> int:
        """The source ID of adjacent edge's network element."""
        pass

    @property
    def soid(self) -> int:
        """The source object ID of adjacent edge's network element."""
        pass

    @property
    def names(self) -> list[SourceFeatureName]:
        """A list of element names referring to the adjacent edge."""
        pass

    @property
    def fromPosition(self) -> float:
        """The position on a network element where the adjacent edge starts.  In the range of 0.0 to 1.0."""
        pass

    @property
    def toPosition(self) -> float:
        """The position on a network element where the adjacent edge ends.  In the range of 0.0 to 1.0."""
        pass

    def getName(self, index: SourceFeatureNameType, name_field: SourceFeatureNameField = SourceFeatureNameField.FullName, fallback_value="") -> str | None:
        """TODO: doc"""
        pass

class SourceFeatureName:
    """
    Represents a set of names that the directions engine may output or uses during analysis.
    If fullName is empty then a name is generated using the values (separated by a space) of the fields:
    prefixDirection, prefixType, baseName, suffixType, suffixDirection, and highwayDirection.
    """

    @property
    def fullName(self) -> str:
        """The full name to display. If it is not populated the name generator will generate a value from existing
        components."""
        pass

    @property
    def baseName(self) -> str:
        """The base name of the element. For example name of a street "Main"."""
        pass

    @property
    def highwayDirection(self) -> str:
        """The highway direction component. For example "N"."""
        pass

    @property
    def prefixType(self) -> str:
        """The type of a base name that goes before the name. For example "St""""
        pass

    @property
    def prefixDirection(self) -> str:
        """The direction prefix of the element name. For example "North"."""
        pass

    @property
    def suffixType(self) -> str:
        """The type of a base name that goes after the name. For example "Ave""""
        pass

    @property
    def suffixDirection(self) -> str:
        """The direction suffix of the element name. For example "North"."""
        pass

    @property
    def phrase(self) -> str:
        """The phase to help generate directions for landmarks. For example "before the 7-Eleven"."""
        pass

    @property
    def language(self) -> str:
        """The language of the elements. For example "en"."""
        pass

    @property
    def nameClass(self) -> NameClass | None:
        """
        The optional class of the name:
        StreetName - Ordinary street name (example: "Lake St" where "Lake" is name and "St" is suffix type)
        RouteNumber - Route number (example: "CR-J22 N" where "CR-J22" is name and "N" is direction)
        """
        pass

class SpatialLandmark:
    """Represents landmark on the network source features when generating directions."""

    @property
    def name(self) -> SourceFeatureName:
        """The name of the landmark."""
        pass

    @property
    def position(self) -> float:
        """The position on a edge. Zero for junctions."""
        pass

    @property
    def distanceInMeters(self) -> float:
        """The distance to the landmark from the route."""
        pass

    @property
    def side(self) -> LandmarkSide:
        """The side of the landmark relative to the route."""
        pass

class ReferenceLandmark:
    """Represents a landmark that features when generating directions."""

    @property
    def importance(self) -> int:
        """The priority of the landmark."""
        pass

    @property
    def position(self) -> float:
        """The position on a edge. Zero for junctions."""
        pass

    @property
    def side(self) -> LandmarkSide:
        """The side of the landmark relative to the route."""
        pass

    @property
    def type(self) -> ReferenceLandmarkType:
        """The reference landmark type."""
        pass

    @property
    def name(self) -> SourceFeatureName:
        """The name of the landmark."""
        pass