# -*- coding: utf-8 -*-
#COPYRIGHT 2020 ESRI
#
#TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
#Unpublished material - all rights reserved under the
#Copyright Laws of the United States.
#
#For additional information, contact:
#Environmental Systems Research Institute, Inc.
#Attn: Contracts Dept
#380 New York Street
#Redlands, California, USA 92373
#
#email: contracts@esri.com
r"""The GeoAI toolbox contains tools for using and training AI models that
work with geospatial and tabular data. These tools use modern machine
learning and deep learning techniques and integrate them with GIS."""
__all__ = ['ClassifyTextUsingDeepLearning', 'ExtractEntitiesUsingDeepLearning',
           'PredictUsingAutoML', 'TrainEntityRecognitionModel',
           'TrainTextClassificationModel', 'TrainTextTransformationModel',
           'TrainUsingAutoML', 'TransformTextUsingDeepLearning']
__alias__ = 'geoai'
from arcpy.geoprocessing._base import gptooldoc, gp, gp_fixargs
from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject

# Feature and Tabular Analysis toolset
@gptooldoc('PredictUsingAutoML_geoai', None)
def PredictUsingAutoML(in_model_definition=None, prediction_type=None, in_features=None, explanatory_rasters=None, distance_features=None, out_prediction_features=None, out_prediction_surface=None, match_explanatory_variables=None, match_distance_variables=None, match_explanatory_rasters=None):
    """PredictUsingAutoML_geoai(in_model_definition, prediction_type, in_features, {explanatory_rasters;explanatory_rasters...}, {distance_features;distance_features...}, {out_prediction_features}, {out_prediction_surface}, {match_explanatory_variables;match_explanatory_variables...}, {match_distance_variables;match_distance_variables...}, {match_explanatory_rasters;match_explanatory_rasters...})

        Uses the trained .dlpk model produced by the Train Using AutoML tool
        to predict continuous variables (regression) or categorical variables
        (classification) on unseen compatible datasets.

     INPUTS:
      in_model_definition (File):
          The .dlpk file or the .emd file.
      prediction_type (String):
          Specifies the output file type that will be created.

          * PREDICT_FEATURE-A feature layer containing the prediction values
          will be created. The Output Prediction Features value is required for
          this option.

          * PREDICT_RASTER-A raster layer containing the prediction values will
          be created. The Output Prediction Surface value is required for this
          option.
      in_features (Feature Layer / Table View / Feature Class):
          The features for which the prediction will be obtained. The input
          should include some or all of the fields necessary to determine the
          dependent variable value. This parameter is required if the
          prediction_type parameter is set to PREDICT_FEATURES.
      explanatory_rasters {Raster Layer}:
          A list of rasters that contain some or all of the fields necessary to
          determine the dependent variable value. This parameter is required if
          the prediction_type parameter is set to PREDICT_RASTER.
      distance_features {Feature Layer}:
          The features whose distances from the input training features will be
          estimated automatically and added as more explanatory variables.
          Distances will be calculated from each of the input explanatory
          training distance features to the nearest input training features. If
          the input explanatory training distance features are polygons or
          lines, the distance attributes will be calculated as the distance
          between the closest segments of the pair of features.
      match_explanatory_variables {Value Table}:
          The mapping of field names from the prediction set to the training
          set. The columns and fields in the test feature class must match the
          field names provided during the training. The string values are the
          column names in the train dataset that match the field names in the
          input feature class. This parameter is only required if the same
          variables are identified by different names in the training and the
          test datasets.
      match_distance_variables {Value Table}:
          The mapping of distance features names from the prediction set to the
          training set. The names of the distance features that were used for
          prediction must match the same distance features used during training.
          The string values are the feature names that were used for prediction
          that match the distance features names used for training. This
          parameter is only required if the same distance features used during
          training and prediction are identified by different names.
      match_explanatory_rasters {Value Table}:
          The mapping of field names from the prediction rasters to the training
          rasters. The names of the explanatory rasters that were used for
          prediction must match the same explanatory rasters used during
          training. The string values are the explanatory raster names that were
          used for prediction that match the explanatory raster names used for
          training. This parameter is only required if the same explanatory
          rasters used during training and prediction are identified by
          different names.

     OUTPUTS:
      out_prediction_features {Feature Class / Table}:
          The output table or feature class.
      out_prediction_surface {Folder}:
          The path to where the output prediction raster will be saved."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.PredictUsingAutoML_geoai(*gp_fixargs((in_model_definition, prediction_type, in_features, explanatory_rasters, distance_features, out_prediction_features, out_prediction_surface, match_explanatory_variables, match_distance_variables, match_explanatory_rasters), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('TrainUsingAutoML_geoai', None)
def TrainUsingAutoML(in_features=None, out_model=None, variable_predict=None, treat_variable_as_categorical=None, explanatory_variables=None, distance_features=None, explanatory_rasters=None, total_time_limit=None, autoML_mode=None, algorithms=None, validation_percent=None, out_report=None, out_importance=None, out_features=None):
    """TrainUsingAutoML_geoai(in_features, out_model, variable_predict, {treat_variable_as_categorical}, {explanatory_variables;explanatory_variables...}, {distance_features;distance_features...}, {explanatory_rasters;explanatory_rasters...}, {total_time_limit}, {autoML_mode}, {algorithms;algorithms...}, {validation_percent}, {out_report}, {out_importance}, {out_features})

        Automates the process and trains a machine learning model to produce a
        deep learning package (.dlpk file). This includes exploratory data
        analysis, feature selection, feature engineering, model selection,
        hyperparameter tuning, and model training. Its outputs include
        performance metrics of the best model on the training data, as well as
        the trained .dlpk model that can be used to predict on a compatible
        new dataset using the Predict Using AutoML tool.

     INPUTS:
      in_features (Feature Layer / Table View / Feature Class):
          The input feature class that will be used to train the model.
      variable_predict (Field):
          A field from the in_features parameter value that contains the values
          that will be used to train the model. This field contains known
          (training) values of the variable that will be used to predict at
          unknown locations.
      treat_variable_as_categorical {Boolean}:
          Specifies whether the variable_predict parameter value will be treated
          as a categorical variable.

          * CATEGORICAL-The variable_predict parameter value will be treated as
          a categorical variable and the tool will perform classification.

          * CONTINUOUS-The variable_predict parameter value will be treated as
          continuous and the tool will perform regression. This is the default.
      explanatory_variables {Value Table}:
          A list of fields representing the explanatory variables that will help
          predict the value or category of the variable_predict parameter value.
          Pass the True value ('name_of_variabler',True) for any variables that
          represent classes or categories (such as land cover, presence, or
          absence).
      distance_features {Feature Layer}:
          The features whose distances from the input training features will be
          estimated automatically and added as more explanatory variables.
          Distances will be calculated from each of the input explanatory
          training distance features to the nearest input training features.
          Point and polygon features are supported, and if the input explanatory
          training distance features are polygons, the distance attributes will
          be calculated as the distance between the closest segments of the pair
          of features.
      explanatory_rasters {Value Table}:
          The rasters whose values will be extracted from the raster and
          considered as explanatory variables for the model. Each layer forms
          one explanatory variable. For each feature in the input training
          features, the value of the raster cell will be extracted at that exact
          location. Bilinear raster resampling will be used when extracting the
          raster value for continuous rasters. Nearest neighbor assignment will
          be used when extracting a raster value from categorical rasters. Check
          the Categorical check box for any raster that represent classes or
          categories such as land cover, presence, or absence.
      total_time_limit {Double}:
          The total time limit in minutes it takes for AutoML model training.
          The default is 60 (1 hour).
      autoML_mode {String}:
          Specifies the goal of AutoML and how intensive the AutoML search will
          be.

          * BASIC-Basic is used to explain the significance of the different
          variables and the data. Feature engineering, feature selection, and
          hyperparameter tuning will not be performed. Full descriptions and
          explanations for model learning curves, feature importance plots
          generated for tree-based models, and SHAP plots for all other models
          will be included in reports. This mode takes the least amount of
          processing time. This is the default.

          * INTERMEDIATE-Intermediate is used to train a model that will be used
          in real-life use cases. It uses 5-fold cross validation (CV) and
          produces output of learning curves and importance plots in the
          reports, but SHAP plots are not available.

          * ADVANCED-Advanced is used for machine learning competitions (for
          maximum performance). This mode uses 10-fold cross validation (CV) and
          performs feature engineering, feature selection, and hyperparameter
          tuning. In this mode, input training features are assigned to multiple
          spatial grids of different sizes based on their location, and the
          corresponding grid IDs are passed as additional categorical
          explanatory variables to the model. The report only includes learning
          curves and model explainability is not available.
      algorithms {String}:
          Specifies the algorithms that will be used during the training.

          * LINEAR-The Linear regression supervised algorithm will be used to
          train a regression machine learning model. In case only linear model
          is chosen then make sure that the total number of records are less
          than 10000 and number of columns less than 1000. Other models can
          handle larger datasets and hence it is suggested to use Linear model
          as one of the many models and not as a standalone model while running
          the tool.

          * RANDOM FOREST-The Random Forest decision tree-based supervised
          machine learning algorithm will be used. It can be used for both
          classification and regression.

          * XGBOOST-The XGBoost (extreme gradient boosting) supervised machine
          learning algorithm will be used. It can be used for both
          classification and regression.

          * LIGHT GBM-The Light GBM gradient boosting ensemble algorithm, which
          is based on decision trees, will be used. It can be used for both
          classification and regression. Light GBM is optimized for high
          performance with distributed systems.

          * DECISION TREE-The Decision Tree supervised machine learning
          algorithm, which classifies or regresses the data using true and false
          answers to certain questions, will be used. Decision trees are easily
          understood and are good for explainability.

          * EXTRA TREE-The Extra Tree (extremely randomized trees) ensemble
          supervised machine learning algorithm, which uses decision trees, will
          be used. This method is similar to Random Forests but can be faster.
          By default, all of the algorithms will be used.
      validation_percent {Long}:
          The percentage of input data that will be used for validation. The
          default value is 10.

     OUTPUTS:
      out_model (File):
          The output trained model that will be saved as a deep learning package
          (.dlpk file).
      out_report {File}:
          Specifies the output report that will be generated as a HTML file. If
          the path provided is not empty then the report will be created inside
          a new folder under the provided path. The report will contain details
          of the various models as well as details of the hyperparameters that
          were used during the evaluation and the performance of each model.
          Hyperparameters are parameters that control the training process. They
          are not updated during training and include model architecture,
          learning rate, number of epochs, and so on.
      out_importance {Table}:
          An output table containing information about the importance of each
          explanatory variable (fields, distance features, and rasters) used in
          the model.
      out_features {Feature Class}:
          The feature layer containing the predicted values by the best
          performing model on the training feature layer. It can be used to
          verify model performance by visually comparing the predicted values
          with the ground truth."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.TrainUsingAutoML_geoai(*gp_fixargs((in_features, out_model, variable_predict, treat_variable_as_categorical, explanatory_variables, distance_features, explanatory_rasters, total_time_limit, autoML_mode, algorithms, validation_percent, out_report, out_importance, out_features), True)))
        return retval
    except Exception as e:
        raise e


# Text Analysis toolset
@gptooldoc('ClassifyTextUsingDeepLearning_geoai', None)
def ClassifyTextUsingDeepLearning(in_table=None, text_field=None, in_model_definition_file=None, class_label_field=None, model_arguments=None):
    """ClassifyTextUsingDeepLearning_geoai(in_table, text_field, in_model_definition_file, {class_label_field}, {model_arguments;model_arguments...})

        Runs a trained text classification model on a text field in a feature
        class or table and updates each record with an assigned class or
        category label.

     INPUTS:
      in_table (Feature Layer / Table View):
          The input point, line, or polygon feature class, or table, containing
          the text that will be classified and labelled.
      text_field (Field):
          A text field in the input feature class or table that contains the
          text that will be classified.
      in_model_definition_file (File):
          The trained model that will be used for classification. The model
          definition file can be an Esri model definition JSON file (.emd) or a
          deep learning model package (.dlpk) stored locally.
      class_label_field {String}:
          The name of the field that will contain the class or category label
          assigned by the model. The default field name is ClassLabel.
      model_arguments {Value Table}:
          Additional arguments, such as a confidence threshold, that will be
          used to adjust the sensitivity of the model.The names of the arguments
          will be populated by the tool."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ClassifyTextUsingDeepLearning_geoai(*gp_fixargs((in_table, text_field, in_model_definition_file, class_label_field, model_arguments), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('ExtractEntitiesUsingDeepLearning_geoai', None)
def ExtractEntitiesUsingDeepLearning(in_folder=None, out_table=None, in_model_definition_file=None, model_arguments=None, batch_size=None, location_zone=None, in_locator=None):
    """ExtractEntitiesUsingDeepLearning_geoai(in_folder, out_table, in_model_definition_file, {model_arguments;model_arguments...}, {batch_size}, {location_zone}, {in_locator})

        Runs a trained named entity recognizer model on text files in a folder
        to extract entities and locations (such as addresses, place or person
        names, dates, and monetary values) in a table. If the extracted
        entities contain an address, the tool geocodes the addresses using the
        specified locator and produces a feature class as an output.

     INPUTS:
      in_folder (Folder):
          The folder containing the text files on which named entity extraction
          will be performed.
      in_model_definition_file (File):
          The trained model that will be used for classification. The model
          definition file can be an Esri model definition JSON file (.emd) or a
          deep learning model package (.dlpk) stored locally.
      model_arguments {Value Table}:
          Additional arguments, such as a confidence threshold, that will be
          used to adjust the sensitivity of the model.The names of the arguments
          will be populated by the tool.
      batch_size {Double}:
          The number of training samples that will be processed at one time. The
          default value is 4.Increasing the batch size can improve tool
          performance; however, as
          the batch size increases, more memory is used. If an out of memory
          error occurs, use a smaller batch size.
      location_zone {String}:
          The geographic region or zone in which the addresses are expected to
          be located. The specified text will be appended to the address
          extracted by the model.The locator uses the location zone information
          to identify the region
          or geographic area in which the address is located and produces better
          results.
      in_locator {Address Locator}:
          The locator that will be used to geocode addresses found in the input
          text documents. A point is generated for each address that is geocoded
          successfully and stored in the output feature class.

     OUTPUTS:
      out_table (Feature Class / Table / Feature Layer):
          The output feature class or table that will contain the extracted
          entities. If a locator is provided and the model extracts addresses,
          the feature class will be produced by geocoding the extracted
          addresses."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.ExtractEntitiesUsingDeepLearning_geoai(*gp_fixargs((in_folder, out_table, in_model_definition_file, model_arguments, batch_size, location_zone, in_locator), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('TrainEntityRecognitionModel_geoai', None)
def TrainEntityRecognitionModel(in_folder=None, out_model=None, pretrained_model_file=None, address_entity=None, max_epochs=None, model_backbone=None, batch_size=None, model_arguments=None, learning_rate=None, validation_percentage=None, stop_training=None, make_trainable=None):
    """TrainEntityRecognitionModel_geoai(in_folder, out_model, {pretrained_model_file}, {address_entity}, {max_epochs}, {model_backbone}, {batch_size}, {model_arguments;model_arguments...}, {learning_rate}, {validation_percentage}, {stop_training}, {make_trainable})

        Trains a named entity recognition model to extract a predefined set of
        entities from raw text.

     INPUTS:
      in_folder (Folder):
          A folder containing training data in the form of standard datasets for
          NER tasks. The training data must be in .json or .csv files. The file
          format determines the dataset type of the input. The following
          are the supported dataset types:

          * ner_json-The training data folder should contain a .json file with
          text and the labelled entities formatted using the spaCy JSON training
          format.

          * IOB-The IOB (I - inside, O - outside, B - beginning tags)
          format proposed by Ramshaw and Marcus in the paper "Text Chunking
          using Transformation-Based Learning". The training data
          folder should contain the following two .csv files:

          * token.csv-Contains text as input chunks.

          * tags.csv-Contains IOB tags for the text chunks.

          * BILUO-An extension of the IOB format that additionally
          contains 'L - last and U - unit tags. The training data
          folder should contain the following two .csv files:

          * token.csv-Contains text as input chunks.

          * tags.csv-Contains BILUO tags for the text chunks.
      pretrained_model_file {File}:
          A pretrained model that will be used to fine-tune the new model. The
          input can be an Esri model definition file (.emd) or a deep learning
          package file (.dlpk).A pretrained model with similar entities can be
          fine-tuned to fit the
          new model. The pretrained model must have been trained with the same
          model type and backbone model that will be used to train the new
          model.
      address_entity {String}:
          An address entity that will be treated as a location. During
          inference, such entities will be geocoded using the specified locator,
          and a feature class will be produced as a result of the entity
          extraction process. If a locator is not provided or the trained model
          does not extract address entities, a table containing the extracted
          entities will be produced instead.
      max_epochs {Long}:
          The maximum number of epochs for which the model will be trained. A
          maximum epoch value of 1 means the dataset will be passed forward and
          backward through the neural network one time. The default value is 5.
      model_backbone {String}:
          Specifies the preconfigured neural network that will be used as the
          architecture for training the new model.

          * bert-base-cased-The model will be trained using the BERT neural
          network. BERT is pretrained using the masked language modeling
          objective and next sentence prediction.

          * roberta-base-The model will be trained using the RoBERTa neural
          network. RoBERTa modifies the key hyperparameters of BERT, eliminating
          the pretraining objective and training of the next sentence with small
          batches and higher learning rates.

          * albert-base-v1-The model will be trained using the ALBERT neural
          network. ALBERT uses a self-supervised loss that focuses on modeling
          intersentence coherence, resulting in better scalability than BERT.

          * xlnet-base-cased-The model will be trained using the XLNet neural
          network. XLNet is a generalized autoregressive pretraining method. It
          allows learning bidirectional contexts by maximizing the expected
          probability on all permutations of the factorization order, which
          overcomes BERT's drawbacks.

          * xlm-roberta-base-The model will be trained using the XLM-RoBERTa
          neural network. XLM-RoBERTa is a multilingual model trained on 100
          different languages. Unlike some XLM multilingual models, it does not
          require language tensors to understand which language is used and
          identifies the correct language from the input IDs.

          * distilroberta-base-DistilRoBERTa is an English language model
          pretrained with the supervision of roberta-base solely on
          OpenWebTextCorpus, a reproduction of OpenAI's WebText dataset.

          * distilbert-base-cased-The model will be trained using the DistilBERT
          neural network. DistilBERT is a smaller general-purpose language
          representation model.

          * spacy-The model will be trained using the SpaCy neural network.
          SpaCy is an open-source library for advanced natural language
          processing.
      batch_size {Double}:
          The number of training samples that will be processed at one time.
          This parameter is not applicable to models with the spaCy backbone.
          The default value is 2.Increasing the batch size can improve tool
          performance; however, as
          the batch size increases, more memory is used. If an out of memory
          error occurs, use a smaller batch size.
      model_arguments {Value Table}:
          Additional arguments for initializing the model, such as seq_len for
          the maximum sequence length of the training data, that will be
          considered for training the model.See keyword arguments in the
          EntityRecognizer documentation for the
          list of supported models arguments that can be used.
      learning_rate {Double}:
          The step size indicating how much the model weights will be adjusted
          during the training process. If no value is specified, an optimal
          learning rate will be deduced automatically.
      validation_percentage {Double}:
          The percentage of training samples that will be used for validating
          the model. The default value is 10.
      stop_training {Boolean}:
          Specifies whether model training will stop when the model is no longer
          improving or until the max_epochs parameter value is reached.

          * STOP_TRAINING-The model training will stop when the model is no
          longer improving, regardless of the max_epochs parameter value
          specified. This is the default.

          * CONTINUE_TRAINING-The model training will continue until the
          max_epochs parameter value is reached.
      make_trainable {Boolean}:
          Specifies whether the backbone layers in the pretrained model will be
          frozen, so that the weights and biases remain as originally designed.

          * TRAIN_MODEL_BACKBONE-The backbone layers will not be frozen, and the
          weights and biases of the model_backbone parameter value can be
          altered to fit the training samples. This takes more time to process
          but typically produces better results. This is the default.

          * FREEZE_MODEL_BACKBONE-The backbone layers will be frozen, and the
          predefined weights and biases of the model_backbone parameter value
          will not be altered during training.

     OUTPUTS:
      out_model (Folder):
          The output folder location that will store the trained model."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.TrainEntityRecognitionModel_geoai(*gp_fixargs((in_folder, out_model, pretrained_model_file, address_entity, max_epochs, model_backbone, batch_size, model_arguments, learning_rate, validation_percentage, stop_training, make_trainable), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('TrainTextClassificationModel_geoai', None)
def TrainTextClassificationModel(in_table=None, text_field=None, label_field=None, out_model=None, pretrained_model_file=None, max_epochs=None, model_backbone=None, batch_size=None, model_arguments=None, learning_rate=None, validation_percentage=None, stop_training=None, make_trainable=None, remove_html_tags=None, remove_urls=None):
    """TrainTextClassificationModel_geoai(in_table, text_field, label_field, out_model, {pretrained_model_file}, {max_epochs}, {model_backbone}, {batch_size}, {model_arguments;model_arguments...}, {learning_rate}, {validation_percentage}, {stop_training}, {make_trainable}, {remove_html_tags}, {remove_urls})

        Trains a text classification model to assign a predefined category or
        label to unstructured text.

     INPUTS:
      in_table (Feature Class / Table View / Feature Layer):
          A feature class or table containing a text field with the input text
          for the model and a label field containing the target class labels.
      text_field (Field):
          A text field in the input feature class or table that contains the
          input text that will be classified by the model.
      label_field (Field):
          A text field in the input feature class or table that contains the
          target class labels for training the model.
      pretrained_model_file {File}:
          A pretrained model that will be used to fine-tune the new model. The
          input can be an Esri model definition file (.emd) or a deep learning
          package file (.dlpk).A pretrained model with similar classes can be
          fine-tuned to fit the
          new model. The pretrained model must have been trained with the same
          model type and backbone model that will be used to train the new
          model.
      max_epochs {Long}:
          The maximum number of epochs for which the model will be trained. A
          maximum epoch value of 1 means the dataset will be passed forward and
          backward through the neural network one time. The default value is 5.
      model_backbone {String}:
          Specifies the preconfigured neural network that will serve as the
          encoder for the model and extract feature representations of the input
          text in the form of fixed length vectors. These vectors are then
          passed as input to the classification head of the model.

          * bert-base-cased-The model will be trained using the BERT neural
          network. BERT is pretrained using the masked language modeling
          objective and next sentence prediction.

          * roberta-base-The model will be trained using the RoBERTa neural
          network. RoBERTa modifies the key hyperparameters of BERT, eliminating
          the pretraining objective and training of the next sentence with small
          batches and higher learning rates.

          * albert-base-v1-The model will be trained using the ALBERT neural
          network. ALBERT uses a self-supervised loss that focuses on modeling
          intersentence coherence, resulting in better scalability than BERT.

          * xlnet-base-cased-The model will be trained using the XLNet neural
          network. XLNet is a generalized autoregressive pretraining method. It
          allows learning bidirectional contexts by maximizing the expected
          probability on all permutations of the factorization order, which
          overcomes BERT's drawbacks.

          * xlm-roberta-base-The model will be trained using the XLM-RoBERTa
          neural network. XLM-RoBERTa is a multilingual model trained on 100
          different languages. Unlike some XLM multilingual models, it does not
          require language tensors to understand which language is used and
          identifies the correct language from the input IDs.

          * distilroberta-base-The model will be trained using the DistilRoBERTa
          neural network. DistilRoBERTa is an English language model pretrained
          with the supervision of roberta-base solely on OpenWebTextCorpus, a
          reproduction of OpenAI's WebText dataset.

          * distilbert-base-cased-The model will be trained using the DistilBERT
          neural network. DistilBERT is a smaller general-purpose language
          representation model.
      batch_size {Double}:
          The number of training samples that will be processed at one time. The
          default value is 2.Increasing the batch size can improve tool
          performance; however, as
          the batch size increases, more memory is used. If an out of memory
          error occurs, use a smaller batch size.
      model_arguments {Value Table}:
          Additional arguments for initializing the model, such as seq_len for
          the maximum sequence length of the training data, that will be
          considered for training the model.See keyword arguments in the
          TextClassifier documentation for the list
          of supported models arguments that can be used.
      learning_rate {Double}:
          The step size indicating how much the model weights will be adjusted
          during the training process. If no value is specified, an optimal
          learning rate will be deduced automatically.
      validation_percentage {Double}:
          The percentage of training samples that will be used for validating
          the model. The default value is 10.
      stop_training {Boolean}:
          Specifies whether model training will stop when the model is no longer
          improving or until the max_epochs parameter value is reached.

          * STOP_TRAINING-The model training will stop when the model is no
          longer improving, regardless of the max_epochs parameter value
          specified. This is the default.

          * CONTINUE_TRAINING-The model training will continue until the
          max_epochs parameter value is reached.
      make_trainable {Boolean}:
          Specifies whether the backbone layers in the pretrained model will be
          frozen, so that the weights and biases remain as originally designed.

          * TRAIN_MODEL_BACKBONE-The backbone layers will not be frozen, and the
          weights and biases of the model_backbone parameter value can be
          altered to fit the training samples. This takes more time to process
          but typically produces better results. This is the default.

          * FREEZE_MODEL_BACKBONE-The backbone layers will be frozen, and the
          predefined weights and biases of the model_backbone parameter value
          will not be altered during training.
      remove_html_tags {Boolean}:
          Specifies whether HTML tags will be removed from the input text.

          * REMOVE_HTML_TAGS-The HTML tags in the input text will be removed.
          This is the default.

          * DO_NOT_REMOVE_HTML_TAGS-The HTML tags in the input text will not be
          removed.
      remove_urls {Boolean}:
          Specifies whether URLs will be removed from the input text.

          * REMOVE_URLS-The URLs in the input text will be removed. This is the
          default.

          * DO_NOT_REMOVE_URLS-The URLs in the input text will not be removed.

     OUTPUTS:
      out_model (Folder):
          The output folder location that will store the trained model."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.TrainTextClassificationModel_geoai(*gp_fixargs((in_table, text_field, label_field, out_model, pretrained_model_file, max_epochs, model_backbone, batch_size, model_arguments, learning_rate, validation_percentage, stop_training, make_trainable, remove_html_tags, remove_urls), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('TrainTextTransformationModel_geoai', None)
def TrainTextTransformationModel(in_table=None, text_field=None, label_field=None, out_model=None, pretrained_model_file=None, max_epochs=None, model_backbone=None, batch_size=None, model_arguments=None, learning_rate=None, validation_percentage=None, stop_training=None, make_trainable=None, remove_html_tags=None, remove_urls=None):
    """TrainTextTransformationModel_geoai(in_table, text_field, label_field, out_model, {pretrained_model_file}, {max_epochs}, {model_backbone}, {batch_size}, {model_arguments;model_arguments...}, {learning_rate}, {validation_percentage}, {stop_training}, {make_trainable}, {remove_html_tags}, {remove_urls})

        Trains a text transformation model to transform, translate, or
        summarize text.

     INPUTS:
      in_table (Feature Class / Table View / Feature Layer):
          A feature class or table containing a text field with the input text
          for the model and a label field containing the target transformed
          text.
      text_field (Field):
          A text field in the input feature class or table that contains the
          input text that will be transformed by the model.
      label_field (Field):
          A text field in the input feature class or table that contains the
          target transformed text for training the model.
      pretrained_model_file {File}:
          A pretrained model that will be used to fine-tune the new model. The
          input can be an Esri model definition file (.emd) or a deep learning
          package file (.dlpk).A pretrained model that performs a similar task
          can be fine-tuned to
          fit the training data. The pretrained model must have been trained
          with the same model type and backbone model that will be used to train
          the new model.
      max_epochs {Long}:
          The maximum number of epochs for which the model will be trained. A
          maximum epoch value of 1 means the dataset will be passed forward and
          backward through the neural network one time. The default value is 5.
      model_backbone {String}:
          Specifies the preconfigured neural network that will be used as the
          architecture for training the new model.

          * t5-small-The new model will be trained using the T5 neural network.
          T5 is a unified framework that converts every language problem into a
          text-to-text format. t5-small is the small variant of T5.

          * t5-base-The new model will be trained using the T5 neural network.
          T5 is a unified framework that converts every language problem into a
          text-to-text format. t5-base is the medium variant of T5.

          * t5-large-The new model will be trained using the T5 neural network.
          T5 is a unified framework that converts every language problem into a
          text-to-text format. t5-large is the large variant of T5.
      batch_size {Double}:
          The number of training samples that will be processed at one time. The
          default value is 2.Increasing the batch size can improve tool
          performance; however, as
          the batch size increases, more memory is used. If an out of memory
          error occurs, use a smaller batch size.
      model_arguments {Value Table}:
          Additional arguments for initializing the model, such as seq_len for
          the maximum sequence length of the training data, that will be
          considered for training the model.See keyword arguments in the
          SequenceToSequence documentation for the
          list of supported models arguments that can be used.
      learning_rate {Double}:
          The step size indicating how much the model weights will be adjusted
          during the training process. If no value is specified, an optimal
          learning rate will be deduced automatically.
      validation_percentage {Double}:
          The percentage of training samples that will be used for validating
          the model. The default value is 10.
      stop_training {Boolean}:
          Specifies whether model training will stop when the model is no longer
          improving or until the max_epochs parameter value is reached.

          * STOP_TRAINING-The model training will stop when the model is no
          longer improving, regardless of the max_epochs parameter value
          specified. This is the default.

          * CONTINUE_TRAINING-The model training will continue until the
          max_epochs parameter value is reached.
      make_trainable {Boolean}:
          Specifies whether the backbone layers in the pretrained model will be
          frozen, so that the weights and biases remain as originally designed.

          * TRAIN_MODEL_BACKBONE-The backbone layers will not be frozen, and the
          weights and biases of the model_backbone parameter value can be
          altered to fit the training samples. This takes more time to process
          but typically produces better results. This is the default.

          * FREEZE_MODEL_BACKBONE-The backbone layers will be frozen, and the
          predefined weights and biases of the model_backbone parameter value
          will not be altered during training.
      remove_html_tags {Boolean}:
          Specifies whether HTML tags will be removed from the input text.

          * REMOVE_HTML_TAGS-The HTML tags in the input text will be removed.
          This is the default.

          * DO_NOT_REMOVE_HTML_TAGS-The HTML tags in the input text will not be
          removed.
      remove_urls {Boolean}:
          Specifies whether URLs will be removed from the input text.

          * REMOVE_URLS-The URLs in the input text will be removed. This is the
          default.

          * DO_NOT_REMOVE_URLS-The URLs in the input text will not be removed.

     OUTPUTS:
      out_model (Folder):
          The output folder location that will store the trained model."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.TrainTextTransformationModel_geoai(*gp_fixargs((in_table, text_field, label_field, out_model, pretrained_model_file, max_epochs, model_backbone, batch_size, model_arguments, learning_rate, validation_percentage, stop_training, make_trainable, remove_html_tags, remove_urls), True)))
        return retval
    except Exception as e:
        raise e

@gptooldoc('TransformTextUsingDeepLearning_geoai', None)
def TransformTextUsingDeepLearning(in_table=None, text_field=None, in_model_definition_file=None, result_field=None, model_arguments=None, batch_size=None, minimum_sequence_length=None, maximum_sequence_length=None):
    """TransformTextUsingDeepLearning_geoai(in_table, text_field, in_model_definition_file, {result_field}, {model_arguments;model_arguments...}, {batch_size}, {minimum_sequence_length}, {maximum_sequence_length})

        Runs a trained sequence-to-sequence model on a text field in a feature
        class or table and updates it with a new field containing the
        converted, transformed, or translated text.

     INPUTS:
      in_table (Feature Layer / Table View / Feature Class):
          The input point, line, or polygon feature class, or table, containing
          the text that will be transformed.
      text_field (Field):
          A text field in the input feature class or table that contains the
          text that will be transformed.
      in_model_definition_file (File):
          The trained model that will be used for classification. The model
          definition file can be an Esri model definition JSON file (.emd) or a
          deep learning model package (.dlpk) stored locally.
      result_field {String}:
          The name of the field that will contain the transformed text in the
          output feature class or table. The default field name is Result.
      model_arguments {Value Table}:
          Additional arguments, such as a confidence threshold, that will be
          used to adjust the sensitivity of the model.The names of the arguments
          will be populated by the tool.
      batch_size {Double}:
          The number of training samples that will be processed at one time. The
          default value is 4.Increasing the batch size can improve tool
          performance; however, as
          the batch size increases, more memory is used. If an out of memory
          error occurs, use a smaller batch size.
      minimum_sequence_length {Double}:
          The minimum number of characters for the output text string. The
          default value is 20.
      maximum_sequence_length {Double}:
          The maximum number of characters for the output text string. The
          default value is 50."""
    from arcpy.geoprocessing._base import gp, gp_fixargs
    from arcpy.arcobjects.arcobjectconversion import convertArcObjectToPythonObject
    try:
        retval = convertArcObjectToPythonObject(gp.TransformTextUsingDeepLearning_geoai(*gp_fixargs((in_table, text_field, in_model_definition_file, result_field, model_arguments, batch_size, minimum_sequence_length, maximum_sequence_length), True)))
        return retval
    except Exception as e:
        raise e


# End of generated toolbox code
del gptooldoc, gp, gp_fixargs, convertArcObjectToPythonObject