# COPYRIGHT 2018 ESRI
#
# TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
# Unpublished material - all rights reserved under the
# Copyright Laws of the United States.
#
# For additional information, contact:
# Environmental Systems Research Institute, Inc.
# Attn: Contracts Dept
# 380 New York Street
# Redlands, California, USA 92373
#
# email: contracts@esri.com

import arcpy


def _is_polygon_feature_class(data):
    desc = arcpy.Describe(data)
    if not _is_feature_class(data) or desc.shapeType != 'Polygon':
        return False
    return True


def _is_feature_class(data):
    desc = arcpy.Describe(data)
    if desc.dataType in ['ShapeFile', 'FeatureClass', 'FeatureService', 'FeatureLayer']:
        return True
    return False


def _is_raster(data):
    try:
        arcpy.Raster(data, is_multidimensional=False)
        return True
    except:
        try:
            arcpy.Raster(data, is_multidimensional=True)
            return True
        except:
            return False


def _get_aoi_from_data_path(data_path):
    if _is_raster(data_path):
        aoi = data_path
    elif _is_polygon_feature_class(data_path):  # not a raster, must be a feature class
        return _get_polygon_from_feature_class(data_path)
    else:
        raise ValueError('{} is neither raster nor polygon feature class'.format(data_path))
    return aoi


def _get_aoi_from_layer(layer):
    if layer.isRasterLayer:
        return layer.dataSource
    elif layer.isFeatureLayer:
        return _get_polygon_from_feature_class(layer)
    else:
        raise ValueError("invalid data layer object")


def _get_polygon_from_feature_class(data):
    geometry = _get_geometry_from_feature_class(data)
    if not isinstance(geometry, arcpy.Polygon):
        raise ValueError("The geometry type is not polygon")
    return geometry


def _get_geometry_from_feature_class(data):
    geometry=None
    cursor = arcpy.da.SearchCursor(data, ["SHAPE@"])
    for row in cursor:
        geometry = geometry.union(row[0]) if geometry else row[0]
    return geometry


def get_aoi(aoi):
    if isinstance(aoi, str):
        aoi = _get_aoi_from_data_path(aoi)
    elif isinstance(aoi, arcpy.Result):  # make layer gp result
        aoi = _get_aoi_from_layer(aoi.getOutput(0))
    elif isinstance(aoi, arcpy.Raster) or \
            isinstance(aoi, arcpy.Extent) or \
            isinstance(aoi, arcpy.Polygon):
        aoi = aoi
    else:
        raise TypeError("invalid parameter: aoi")

    return aoi

def get_geometry(data):
    if data is None:
        return None
    if isinstance(data, str):
        if _is_raster(data):
            return arcpy.Raster(data).extent
        elif _is_feature_class(data):
            return _get_geometry_from_feature_class(data)
        else:
            raise ValueError('can\'t get geometry from data')
    elif isinstance(data, arcpy.Result):
        output = data.getOutput(0)
        if output.isRasterLayer:
            extent=arcpy.Raster(output.dataSource).extent
            return extent
        elif output.isFeatureLayer:
            return _get_geometry_from_feature_class(output)
        else:
            raise ValueError('invalid data layer object')
    elif isinstance(data, arcpy.Geometry) or isinstance(data, arcpy.Extent):
        return data
    else:
        raise ValueError("can\'t get geometry from data")
